import shutil
from sklearn.model_selection import train_test_split
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p as ensure_dir_exists, subdirs
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import get_raw_data_path, get_source_path


def convert(path: str = get_source_path(), subdir: str = "MBAS_Dataset"):
    # INPUT DATA
    path = f"{path}/{subdir}"

    # Train/Test Splits
    images_dir = join(path, "Training")
    training_samples, test_samples = train_test_split(subdirs(images_dir, join=False), random_state=35219)

    ###OUTPUT DATA
    # Target names
    task_name = "Task014_MBAS24"
    task_prefix = "MBAS24"

    # Target paths
    target_base = join(get_raw_data_path(), task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    ensure_dir_exists(target_imagesTr)
    ensure_dir_exists(target_labelsTs)
    ensure_dir_exists(target_imagesTs)
    ensure_dir_exists(target_labelsTr)

    for sTr in training_samples:
        src_image_file_path = join(images_dir, sTr, sTr + "_gt.nii.gz")
        src_label_file_path = join(images_dir, sTr, sTr + "_label.nii.gz")
        dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz"

        dst_label_path = f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz"
        shutil.copy2(src_image_file_path, dst_image_file_path)
        shutil.copy2(src_label_file_path, dst_label_path)

    del sTr
    for sTs in test_samples:
        src_image_file_path = join(images_dir, sTs, sTs + "_gt.nii.gz")
        src_label_file_path = join(images_dir, sTs, sTs + "_label.nii.gz")
        dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz"

        dst_label_path = f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz"
        shutil.copy2(src_image_file_path, dst_image_file_path)
        shutil.copy2(src_label_file_path, dst_label_path)

    del sTs
    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("MRI (LGE)",),
        labels={0: "background", 1: "Right Atrium Cavity", 2: "Left Atrium Cavity", 3: "Left & Right Atrium Wall"},
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="MICCAI MBAS 2024",
        dataset_reference="https://codalab.lisn.upsaclay.fr/competitions/18516#learn_the_details",
    )


if __name__ == "__main__":
    convert()



from torch import Tensor, argmax, sigmoid
from torch import nn
from torchmetrics.classification import MulticlassF1Score
from torchmetrics.classification import Accuracy as TorchAccuracy
from torchmetrics.classification import AUROC as TorchAUROC
from torchmetrics.segmentation import GeneralizedDiceScore as TorchGeneralizedDiceScore
from torchmetrics import Metric


class F1(MulticlassF1Score):
    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return super().forward(input, target)


class AUROC(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.method = TorchAUROC(*args, **kwargs)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return self.method(input, target.long())

    def compute(self):
        pass

    def update(self):
        pass


class Accuracy(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.method = TorchAccuracy(*args, **kwargs)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return self.method(input, target.long())

    def compute(self):
        pass

    def update(self):
        pass


class GeneralizedDiceScore(Metric):
    def __init__(self, multilabel: bool, num_classes: int, per_class: bool = False, average: bool = False, *args, **kwargs):
        super().__init__()
        self.multilabel = multilabel
        self.num_classes = num_classes

        self.per_class = per_class
        self.average = average

        self.method = TorchGeneralizedDiceScore(
            num_classes=self.num_classes, per_class=self.per_class or self.average, *args, **kwargs
        )

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if self.multilabel:
            input = (sigmoid(input) > 0.5).long()
        else:
            # GeneralizedDice doesn't support logits, so we need to argmax it.
            input = argmax(input, dim=1)
            target = target.squeeze()

            # The following is due to a bug in torchmetrics. This can be removed when GeneralizedDiceScore has a property
            # argument `input_type` which is currently on their main branch, but released yet.
            # We should then set `input_type=one_hot` only when data is from regions.
            input = nn.functional.one_hot(input, num_classes=self.num_classes).movedim(-1, 1)
            target = nn.functional.one_hot(target.long(), num_classes=self.num_classes).movedim(-1, 1)
            assert len(input.shape) <= 5, input.shape
            assert len(target.shape) <= 5, target.shape

        output = self.method(input, target)

        if self.average:
            return output.mean()

        return output

    def compute(self):
        pass

    def update(self):
        pass



from yucca.pipeline.planning.YuccaPlanner import UnsupervisedPlanner


class UnsupervisedPlannerUnitSpacing(UnsupervisedPlanner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = str(self.__class__.__name__)
        self.preprocessor = "UnsupervisedPreprocessor"  # hard coded

    def determine_target_size_from_fixed_size_or_spacing(self):
        self.fixed_target_spacing = [1.0, 1.0, 1.0]
        self.fixed_target_size = None



from yucca.pipeline.planning.YuccaPlanner import YuccaPlanner


class YuccaPlanner_NoResample(YuccaPlanner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = str(self.__class__.__name__)

    def determine_target_size_from_fixed_size_or_spacing(self):
        self.fixed_target_spacing = None
        self.fixed_target_size = None



import torch
from torch import nn
from yucca.modules.optimization.loss_functions.CE import CE
from yucca.modules.optimization.loss_functions.nnUNet_losses import SoftDiceLoss, sum_tensor
import torch.nn.functional as F


class SoftSkeletonRecallLoss(nn.Module):
    def __init__(self, apply_softmax=True, batch_dice=False, do_bg=True, smooth=1.0):
        super(SoftSkeletonRecallLoss, self).__init__()

        self.do_bg = do_bg
        self.batch_dice = batch_dice
        self.apply_softmax = apply_softmax
        self.smooth = smooth

    def forward(self, x, y, loss_mask=None):
        shp_x, shp_y = x.shape, y.shape

        if self.batch_dice:
            axes = [0] + list(range(2, len(shp_x)))
        else:
            axes = list(range(2, len(shp_x)))

        if self.apply_softmax is True:
            x = F.softmax(x)

        x = x[:, 1:]

        with torch.no_grad():
            if len(shp_x) != len(shp_y):
                y = y.view((shp_y[0], 1, *shp_y[1:]))

            if all([i == j for i, j in zip(shp_x, shp_y)]):
                y_onehot = y[:, 1:]
            else:
                gt = y.long()
                y_onehot = torch.zeros(shp_x, device=x.device, dtype=y.dtype)
                y_onehot.scatter_(1, gt, 1)
                y_onehot = y_onehot[:, 1:]

            sum_gt = sum_tensor(y_onehot, axes) if loss_mask is None else sum_tensor(y_onehot * loss_mask, axes)

        inter_rec = sum_tensor((x * y_onehot), axes) if loss_mask is None else sum_tensor(x * y_onehot * loss_mask, axes)

        rec = (inter_rec + self.smooth) / (sum_gt + self.smooth)

        rec = rec.mean()
        return -rec


class DC_SkelREC_and_CE_loss(nn.Module):
    def __init__(
        self,
        soft_dice_kwargs={},
        soft_skelrec_kwargs={},
        ce_kwargs={},
        weight_ce=1,
        weight_dice=1,
        weight_srec=1,
        log_dice=False,
        ignore_label=None,
    ):
        """
        Weights for CE and Dice do not need to sum to one. You can set whatever you want.
        :param soft_dice_kwargs:
        :param soft_skelrec_kwargs:
        :param ce_kwargs:
        :param aggregate:
        :param square_dice:
        :param weight_ce:
        :param weight_dice:
        """
        super(DC_SkelREC_and_CE_loss, self).__init__()

        if ignore_label is not None:
            ce_kwargs["reduction"] = "none"
        self.log_dice = log_dice
        self.weight_dice = weight_dice
        self.weight_ce = weight_ce
        self.weight_srec = weight_srec
        self.ignore_label = ignore_label

        self.ce = CE()
        self.dc = SoftDiceLoss(**soft_dice_kwargs)
        self.srec = SoftSkeletonRecallLoss(**soft_skelrec_kwargs)

    def forward(self, net_output, target, skel):
        """
        target must be b, c, x, y(, z) with c=1
        :param net_output:
        :param target:
        :return:
        """
        if self.ignore_label is not None:
            assert target.shape[1] == 1, "not implemented for one hot encoding"
            mask = target != self.ignore_label
            target[~mask] = 0
            mask = mask.float()
        else:
            mask = None

        dc_loss = self.dc(net_output, target, loss_mask=mask) if self.weight_dice != 0 else 0
        if self.log_dice:
            dc_loss = -torch.log(-dc_loss)

        ce_loss = self.ce(net_output, target[:, 0].long()) if self.weight_ce != 0 else 0

        if self.ignore_label is not None:
            ce_loss *= mask[:, 0]
            ce_loss = ce_loss.sum() / mask.sum()

        srec_loss = self.srec(net_output, skel) if self.weight_srec != 0 else 0

        result = self.weight_ce * ce_loss + self.weight_dice * dc_loss + self.weight_srec * srec_loss
        return result



import numpy as np
import torch
import torch.nn.functional as F


def crop_to_box(array, bbox):
    """
    Crops an array to the Bounding Box indices
    Should be a list of [xmin, xmax, ymin, ymax (, zmin, zmax)]

    We add +1 because slicing excludes the high val index. Which it should not do here.
    """
    if len(bbox) > 5:
        bbox_slices = (
            slice(bbox[0], bbox[1] + 1),
            slice(bbox[2], bbox[3] + 1),
            slice(bbox[4], bbox[5] + 1),
        )
    else:
        bbox_slices = (slice(bbox[0], bbox[1] + 1), slice(bbox[2], bbox[3] + 1))
    return array[bbox_slices]


def pad_to_size(array, size, **kwargs):
    pad_box = get_pad_box(array, size)
    if len(pad_box) > 5:
        array_padded = np.pad(
            array,
            (
                (pad_box[0], pad_box[1]),
                (pad_box[2], pad_box[3]),
                (pad_box[4], pad_box[5]),
            ),
            **kwargs,
        )
        return array_padded, pad_box

    array_padded = np.pad(
        array,
        ((pad_box[0], pad_box[1]), (pad_box[2], pad_box[3])),
        **kwargs,
    )
    return array_padded, pad_box


def get_pad_box(original_array, min_size):
    assert len(original_array.shape) in [2, 3], "incorrect array shape"
    assert len(min_size) in [2, 3], "incorrect patch_size shape"
    pad_box = []

    if len(original_array.shape) == 3:
        if len(min_size) == 2:
            # 3D Data for 2D model.
            min_size = (0, *min_size)
        for i in range(3):
            val = max(0, min_size[i] - original_array.shape[i])
            pad_box.append(val // 2)
            pad_box.append(val // 2 + val % 2)
        return pad_box

    # 2D data + 2D model
    for i in range(2):
        val = max(0, min_size[i] - original_array.shape[i])
        pad_box.append(val // 2)
        pad_box.append(val // 2 + val % 2)
    return pad_box


def get_pad_kwargs(data, pad_value):
    if pad_value == "min":
        pad_kwargs = {"constant_values": data.min(), "mode": "constant"}
    elif pad_value == "zero":
        pad_kwargs = {"constant_values": np.zeros(1, dtype=data.dtype), "mode": "constant"}
    elif isinstance(pad_value, int) or isinstance(pad_value, float):
        pad_kwargs = {"constant_values": pad_value, "mode": "constant"}
    elif pad_value == "edge":
        pad_kwargs = {"mode": "edge"}
    else:
        print("Unrecognized pad value detected.")
    return pad_kwargs


def ensure_batch_fits_patch_size(batch, patch_size):
    """
    Pads the spatial dimensions of the input tensor so that they are at least the size of the patch dimensions.
    If all spatial dimensions are already larger than or equal to the patch size, the input tensor is returned unchanged.

    Parameters:
    - batch: dict
        a dict with keys {"data": data, "data_properties": data_properties, "case_id": case_id},
        where data is a Tensor of shape (B, C, *spatial_dims)

    - patch_size: tuple of ints
        The minimum desired size for each spatial dimension.

    Returns:
    - padded_input: torch.Tensor
        The input tensor padded to the desired spatial dimensions.
    """
    image = batch["data"]
    image_properties = batch["data_properties"]
    spatial_dims = image.dim() - 2  # Subtract batch and channel dimensions

    if spatial_dims == len(patch_size) + 1:  # 2D patches from 3D data
        patch_size = (1,) + patch_size

    if spatial_dims != len(patch_size):
        raise ValueError(
            f"Input spatial dimensions and patch size dimensions do not match. Got patch_size: {patch_size} and spatial_dims: {spatial_dims} from image of shape: {image.shape}"
        )

    current_sizes = image.shape[2:]  # Spatial dimensions

    current_sizes_tensor = torch.tensor(current_sizes)
    patch_size_tensor = torch.tensor(patch_size)

    pad_sizes = torch.clamp(patch_size_tensor - current_sizes_tensor, min=0)
    pad_left = pad_sizes // 2
    pad_right = pad_sizes - pad_left

    # Construct padding tuple in reverse order for F.pad
    padding_reversed = []
    for left, right in zip(reversed(pad_left.tolist()), reversed(pad_right.tolist())):
        padding_reversed.extend([left, right])

    padded_input = F.pad(image, padding_reversed)

    image_properties["padded_shape"] = np.array(padded_input.shape[2:])
    image_properties["padding"] = list(reversed(padding_reversed))

    return padded_input, image_properties



import torch


def torch_mask(image: torch.Tensor, pixel_value: float, ratio: float, token_size: list[int]) -> torch.Tensor:
    """
    We need to mask image over all channels thus input should be 4d tensor of shape (x, y, z) or 3d tensor of shape (x, y)
    """

    input_shape = image.shape

    if len(token_size) == 1:
        token_size *= len(input_shape)
    assert len(input_shape) == len(
        token_size
    ), f"mask token size not compatible with input data — token: {token_size}, image shape: {input_shape}"

    input_shape_tensor = image.new_tensor(input_shape, dtype=torch.int)
    token_size_tensor = image.new_tensor(token_size, dtype=torch.int)
    grid_dims = torch.ceil(input_shape_tensor / token_size_tensor).to(dtype=torch.int)
    grid_size = torch.prod(grid_dims).item()

    grid_flat = image.new_ones(grid_size)

    grid_flat[: int(grid_size * ratio)] = 0
    grid_flat = grid_flat[torch.randperm(grid_size, device=image.device)]

    grid = grid_flat.view(*grid_dims)

    for dim, size in enumerate(token_size):
        grid = grid.repeat_interleave(size, dim=dim)

    slices = tuple(slice(0, s) for s in input_shape)
    mask = grid[slices]

    image[mask == 0] = pixel_value
    return image, mask


def torch_mask_all_channels(image: torch.Tensor, pixel_value: float, ratio: float, token_size: list[int]) -> torch.Tensor:
    """
    Expects (c,x,y,z) or (c,x,y) as input.
    """
    image[0], mask = torch_mask(image=image[0], pixel_value=pixel_value, ratio=ratio, token_size=token_size)
    for i in range(1, image.shape[0]):
        image[i], _ = torch_mask(image=image[i], pixel_value=pixel_value, ratio=ratio, token_size=token_size)
    return image, mask



import torch
import torchvision.transforms.functional as TF


def torch_blur(image: torch.tensor, sigma: float, clip_to_input_range: bool = True) -> torch.tensor:
    img_min = image.min()
    img_max = image.max()

    if image.ndim == 2:
        image = blur_2D(image, sigma)
    elif image.ndim == 3:
        image = blur_3D(image, sigma)
    else:
        raise ValueError(f"Unsupported image shape for blur: {image.shape}")

    if clip_to_input_range:
        image = torch.clamp(image, min=img_min, max=img_max)
    return image


def blur_2D(image: torch.tensor, sigma: float) -> torch.tensor:
    assert image.ndim == 2, "Expected [H, W] tensor"

    image = image.unsqueeze(0).unsqueeze(0)
    image = TF.gaussian_blur(image, kernel_size=int(2 * round(3 * sigma) + 1), sigma=sigma)
    return image.squeeze(0).squeeze(0)


def blur_3D(image: torch.Tensor, sigma: float) -> torch.Tensor:
    assert image.ndim == 3, "Expected [D, H, W] tensor"

    kernel_size = int(2 * round(3 * sigma) + 1)
    if kernel_size % 2 == 0:
        kernel_size += 1

    # : filter single-channel, [out_channels, in_channels, kernel_size]
    coords = torch.arange(kernel_size) - kernel_size // 2
    kernel = torch.exp(-(coords.float() ** 2) / (2 * sigma**2))
    kernel = kernel / kernel.sum()
    kernel = kernel.view(1, 1, -1)

    # the permutation ordering is moving the target spatial dimension slice to the last dimension since
    # this is where conv1d applies the filter.
    volume = image
    for axis in range(3):
        permute_order = list(range(3))
        permute_order[axis], permute_order[-1] = permute_order[-1], permute_order[axis]
        volume = volume.permute(permute_order).contiguous()

        shape = volume.shape
        volume = volume.view(-1, 1, shape[-1])
        volume = torch.nn.functional.conv1d(volume, kernel.to(volume.device), padding=kernel.shape[-1] // 2)
        volume = volume.view(shape)
        volume = volume.permute(permute_order).contiguous()

    return volume



import subprocess
import os
import yucca


def test_task_convert():
    subprocess.run(["yucca_convert_task", "-t", "Task000_TEST_SEGMENTATION"], check=True)


def test_functional_preprocessing():
    print(yucca.__path__[0])
    subprocess.run(
        ["python", os.path.join(yucca.__path__[0], "documentation/templates/functional_preprocessing.py")], check=True
    )


def test_functional_training():
    subprocess.run(["python", os.path.join(yucca.__path__[0], "documentation/templates/functional_training.py")], check=True)


def test_functional_inference():
    subprocess.run(["python", os.path.join(yucca.__path__[0], "documentation/templates/functional_inference.py")], check=True)



if __name__ == "__main__":
    import os
    import numpy as np
    from PIL import Image
    from batchgenerators.utilities.file_and_folder_operations import (
        maybe_mkdir_p,
        join,
        subfiles,
        save_json,
    )
    from yucca.pipeline.task_conversion.utils import generate_dataset_json
    from yucca.paths import get_raw_data_path, get_preprocessed_data_path

    """
    FASS nnU-Net task conversion script.
    """

    base = "/zhome/af/0/210164/data/FASS/task_converted"
    file_extension = ".png"
    target_dataset_id = 300
    target_dataset_name = f"Task{target_dataset_id:3.0f}_FASS"

    raw_dir = get_raw_data_path()
    maybe_mkdir_p(join(raw_dir, target_dataset_name))

    target_imagesTr = join(raw_dir, target_dataset_name, "imagesTr")
    target_labelsTr = join(raw_dir, target_dataset_name, "labelsTr")
    target_imagesTs = join(raw_dir, target_dataset_name, "imagesTs")
    target_labelsTs = join(raw_dir, target_dataset_name, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTr)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTs)

    tr_cases = subfiles(os.path.join(base, "train", "images"), join=False)
    val_cases = subfiles(os.path.join(base, "val", "images"), join=False)
    test_cases = subfiles(os.path.join(base, "test", "images"), join=False)

    tr_split = []
    val_split = []

    for case in tr_cases:
        case = case[: -len(file_extension)]

        image = Image.open(join(base, "train", "images", case + file_extension))
        image = np.array(image)[:, :, 0]
        image = Image.fromarray(image)
        image.save(join(target_imagesTr, case + "_000.png"))

        label = Image.open(join(base, "train", "annotations", case + file_extension))
        label.save(join(target_labelsTr, case + ".png"))

        tr_split.append(case)

    for case in val_cases:
        case = case[: -len(file_extension)]

        image = Image.open(join(base, "val", "images", case + file_extension))
        image = np.array(image)[:, :, 0]
        image = Image.fromarray(image)
        image.save(join(target_imagesTr, case + "_000.png"))

        label = Image.open(join(base, "val", "annotations", case + file_extension))
        label.save(join(target_labelsTr, case + ".png"))

        val_split.append(case)

    for case in test_cases:
        case = case[: -len(file_extension)]

        image = Image.open(join(base, "test", "images", case + file_extension))
        image = np.array(image)[:, :, 0]
        image = Image.fromarray(image)
        image.save(join(target_imagesTs, case + "_000.png"))

        label = Image.open(join(base, "test", "annotations", case + file_extension))
        label.save(join(target_labelsTs, case + ".png"))

    # manual splits without extensions or modality encoding
    splits = [{"train": tr_split, "val": val_split}]

    pp_out_dir = join(get_preprocessed_data_path(), target_dataset_name)
    maybe_mkdir_p(pp_out_dir)
    save_json(splits, join(pp_out_dir, "splits_final.json"), sort_keys=False)

    labels = {
        0: "background",
        1: "artery",
        2: "liver",
        3: "stomach",
        4: "vein",
    }

    generate_dataset_json(
        join(raw_dir, target_dataset_name, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("Gray",),
        labels=labels,
        dataset_name=target_dataset_name,
    )



import torch


def torch_additive_noise(image, mean, sigma, clip_to_input_range: bool = False):
    # J = I+n
    img_min = image.min()
    img_max = image.max()
    image = image + torch.normal(mean, sigma, image.shape, device=image.device)
    if clip_to_input_range:
        image = torch.clamp(image, min=img_min, max=img_max)
    return image


def torch_multiplicative_noise(image, mean, sigma, clip_to_input_range: bool = False):
    # J = I + I*n
    img_min = image.min()
    img_max = image.max()
    gauss = torch.normal(mean, sigma, image.shape, device=image.device)
    image = image + image * gauss
    if clip_to_input_range:
        image = torch.clamp(image, min=img_min, max=img_max)
    return image



import torch


def torch_bias_field(image: torch.Tensor, clip_to_input_range: bool = False) -> torch.Tensor:
    device = image.device
    img_min = image.min()
    img_max = image.max()

    if len(image.shape) == 3:
        assert image.ndim == 3, "Expected [H, W, D] tensor"
        x, y, z = image.shape
        X, Y, Z = torch.meshgrid(
            torch.linspace(0, x - 1, x, device=device),
            torch.linspace(0, y - 1, y, device=device),
            torch.linspace(0, z - 1, z, device=device),
            indexing="ij",
        )
        x0 = torch.randint(0, x, (1,), device=device)
        y0 = torch.randint(0, y, (1,), device=device)
        z0 = torch.randint(0, z, (1,), device=device)
        G = 1 - ((X - x0) ** 2 / (x**2) + (Y - y0) ** 2 / (y**2) + (Z - z0) ** 2 / (z**2))
    else:
        assert image.ndim == 2, "Expected [H, W] tensor"

        x, y = image.shape
        X, Y = torch.meshgrid(
            torch.linspace(0, x - 1, x, device=device), torch.linspace(0, y - 1, y, device=device), indexing="ij"
        )
        x0 = torch.randint(0, x, (1,), device=device)
        y0 = torch.randint(0, y, (1,), device=device)
        G = 1 - ((X - x0) ** 2 / (x**2) + (Y - y0) ** 2 / (y**2))
    image = G * image

    if clip_to_input_range:
        image = torch.clamp(image, min=img_min, max=img_max)

    return image



import torch
import torch.nn.functional as F
from typing import Optional
from yucca.functional.array_operations.matrix_ops import Rx, Ry, Rz, Rz2D


def _create_zero_centered_coordinate_matrix(shape: tuple[int, ...]) -> torch.Tensor:
    # Using standard numpy indexing 'ij', 2D: (H, W) = (y, x), 3D: (D, H, W) = (z, y, x)
    mesh = torch.stack(torch.meshgrid(*[torch.arange(s, dtype=torch.float32) for s in shape], indexing="ij"))
    for d, s in enumerate(shape):
        mesh[d] -= (s - 1) / 2.0
    return mesh


def torch_spatial(
    image: torch.Tensor,
    patch_size: tuple[int],
    p_deform: float,
    p_rot: float,
    p_rot_per_axis: float,
    p_scale: float,
    alpha: float,
    sigma: float,
    x_rot: float,
    y_rot: float,
    z_rot: float,
    scale_factor: float,
    clip_to_input_range: bool,
    label: Optional[torch.Tensor] = None,
    skip_label: bool = False,
    do_crop: bool = True,
    random_crop: bool = True,
    interpolation_mode: str = "bilinear",
    seed: Optional[int] = None,
):
    if seed is not None:
        torch.manual_seed(seed)

    device, dtype, orig_ndim = image.device, image.dtype, image.ndim
    ndim = len(patch_size)

    # Expand dims if needed
    def _prepare(x):
        return x[None, None] if orig_ndim == ndim else x[None] if orig_ndim == ndim + 1 else x

    image = _prepare(image)
    if label is not None:
        label = _prepare(label)

    if not do_crop:
        patch_size = image.shape[2:]

    coords = _create_zero_centered_coordinate_matrix(patch_size).to(device, dtype)
    if torch.rand(1) < p_deform:
        if ndim == 2:
            # Separable 2D blur
            noise = torch.randn(1, 1, *patch_size, device=device, dtype=dtype)
            ksize = 21
            ax = torch.arange(ksize, device=device, dtype=dtype) - ksize // 2
            k = torch.exp(-0.5 * (ax / sigma) ** 2)
            k /= k.sum()
            ky = k.view(1, 1, -1, 1)
            kx = k.view(1, 1, 1, -1)
            noise = F.conv2d(noise, ky, padding=(ksize // 2, 0), groups=1)
            noise = F.conv2d(noise, kx, padding=(0, ksize // 2), groups=1)
        else:
            # Separable 3D blur
            noise = torch.randn(1, ndim, *patch_size, device=device, dtype=dtype)
            ksize = 9
            ax = torch.arange(ksize, device=device, dtype=dtype) - ksize // 2
            k = torch.exp(-0.5 * (ax / sigma) ** 2)
            k /= k.sum()
            for dim in range(3):
                shape = [1, 1, 1, 1, 1]
                shape[dim + 2] = ksize
                kernel = k.view(*shape).repeat(ndim, 1, 1, 1, 1)
                padding = [ksize // 2 if i == dim else 0 for i in range(3)]
                noise = F.conv3d(noise, kernel, padding=padding, groups=ndim)
        coords += noise[0] * alpha

    if torch.rand(1) < p_rot:
        rot = torch.eye(ndim, device=device, dtype=dtype)
        if ndim == 2:
            rot = rot @ torch.from_numpy(Rz2D(z_rot)).to(device=device, dtype=dtype)
        else:
            if torch.rand(1) < p_rot_per_axis:
                rot = rot @ torch.from_numpy(Rx(x_rot)).to(device=device, dtype=dtype)
            if torch.rand(1) < p_rot_per_axis:
                rot = rot @ torch.from_numpy(Ry(y_rot)).to(device=device, dtype=dtype)
            if torch.rand(1) < p_rot_per_axis:
                rot = rot @ torch.from_numpy(Rz(z_rot)).to(device=device, dtype=dtype)
        coords = (rot @ coords.view(ndim, -1)).view_as(coords)

    if torch.rand(1) < p_scale:
        coords *= scale_factor

    if random_crop and do_crop:
        for d in range(ndim):
            lo = patch_size[d] // 2
            hi = image.shape[d + 2] - patch_size[d] // 2 + 1
            coords[d] += torch.randint(lo, hi, (1,), device=device)
    else:
        for d in range(ndim):
            coords[d] += image.shape[d + 2] / 2 - 0.5

    for d in range(ndim):
        coords[d] = 2 * coords[d] / (image.shape[d + 2] - 1) - 1

    # Swap axes to (x, y) or (x, y, z) order for grid_sample (torch does not default to numpy indexing here)
    grid = coords.permute(*range(1, ndim + 1), 0)
    grid = torch.stack([grid] * image.shape[0], dim=0)

    if ndim == 2:
        grid = grid[..., [1, 0]]
    elif ndim == 3:
        grid = grid[..., [2, 1, 0]]
    else:
        raise ValueError("Only 2D and 3D supported")
    grid_sample_args = {"mode": interpolation_mode, "padding_mode": "zeros", "align_corners": True}
    image_canvas = F.grid_sample(image, grid, **grid_sample_args)
    if clip_to_input_range:
        image_canvas = torch.clamp(image_canvas, min=image.min(), max=image.max())

    if label is not None and not skip_label:
        label_canvas = F.grid_sample(label.float(), grid, mode="nearest", padding_mode="zeros", align_corners=True)
        label_canvas = label_canvas.to(label.dtype)
    else:
        label_canvas = None

    def _restore(x: Optional[torch.Tensor], target_ndim: int) -> Optional[torch.Tensor]:
        if x is None:
            return None
        # For 3D volumes, we want to keep the spatial dimensions
        current_dims = x.ndim
        for i in range(current_dims - target_ndim):
            x = x.squeeze(0)
        return x

    return _restore(image_canvas, orig_ndim), (
        _restore(label_canvas, orig_ndim if label is not None else 0) if label_canvas is not None else None
    )


if __name__ == "__main__":
    import torch
    import matplotlib.pyplot as plt
    import numpy as np

    array = torch.zeros((1, 12, 12, 12))
    array[:, 4:8, 4:8, 4:8] = 1
    array[:, 4, 4:8, 4:8] = 2
    array = torch.from_numpy(
        np.load("/Users/zcr545/Desktop/Projects/repos/asparagus_data/preprocessed_data/Task001_OASIS/imagesTr/1000.nii.npy")
    ).squeeze(0)
    array = torch.stack([array, array], dim=0)
    print(array.shape)
    array = torch.stack([array, array, array])
    print(array.shape)

    out, _ = torch_spatial(
        array,
        patch_size=array.shape[-3:],
        p_deform=0,
        p_rot=1,
        p_rot_per_axis=1,
        p_scale=0,
        alpha=5,  # 5-20
        sigma=10,  # 5-20
        x_rot=0,
        y_rot=0,
        z_rot=0.1,
        scale_factor=0,
        clip_to_input_range=False,
        do_crop=False,
    )
    print("IN", array.shape, "OUT", out.shape)
    fig, ax = plt.subplots(1, 2)
    ax[0].imshow(array[0, :, 40], cmap="gray")
    ax[1].imshow(out[:, 40], cmap="gray")



from yucca.pipeline.managers.YuccaManager import YuccaManager
from yucca.modules.data.augmentation.augmentation_presets import generic
from yucca.modules.lightning_modules.YuccaLightningModule_skeleton_loss import YuccaLightningModule_skeleton_loss


class YuccaManager_SkeletonLoss(YuccaManager):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.lightning_module = YuccaLightningModule_skeleton_loss
        self.loss = "DC_SkelREC_and_CE_loss"
        self.augmentation_params = generic
        self.augmentation_params["skeleton"] = True
        self.augmentation_params["do_tubes"] = True



import wandb
import torch
from yucca.modules.lightning_modules.YuccaLightningModule import YuccaLightningModule
from yucca.modules.optimization.loss_functions.Skeleton_recall_loss import DC_SkelREC_and_CE_loss


class YuccaLightningModule_skeleton_loss(YuccaLightningModule):
    def __init__(
        self,
        config,
        *args,
        **kwargs,
    ):
        super().__init__(
            config=config,
            *args,
            **kwargs,
        )
        self.loss_fn = DC_SkelREC_and_CE_loss

    def training_step(self, batch, batch_idx):
        inputs, target, file_path, skel = batch["image"], batch["label"], batch["file_path"], batch["skel"]
        output = self(inputs)
        loss = self.loss_fn_val(output, target, skel)

        if self.deep_supervision:
            # If deep_supervision is enabled output and target will be a list of (downsampled) tensors.
            # We only need the original ground truth and its corresponding prediction which is always the first entry in each list.
            output = output[0]
            target = target[0]

        metrics = self.compute_metrics(self.train_metrics, output, target)
        self.log_dict(
            {"train/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and wandb.run is not None and self.log_image_this_epoch is True:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="train",
                task_type=self.task_type,
            )

        return loss

    def validation_step(self, batch, batch_idx):
        inputs, target, file_path, skel = batch["image"], batch["label"], batch["file_path"], batch["skel"]
        output = self(inputs)

        loss = self.loss_fn_val(output, target, skel)

        metrics = self.compute_metrics(self.val_metrics, output, target)
        self.log_dict(
            {"val/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and wandb.run is not None and self.log_image_this_epoch is True:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="val",
                task_type=self.task_type,
            )



"""
PLEASE READ YUCCA/DOCUMENTATION/TUTORIALS/ENVIRONMENT_VARIABLES.MD FOR INFORMATION ON HOW TO SET THIS UP
"""

import os
from dotenv import load_dotenv

from batchgenerators.utilities.file_and_folder_operations import maybe_mkdir_p as ensure_dir_exists


def var_is_set(var):
    return var in os.environ.keys()


def get_environment_variable(var):
    load_dotenv()
    if not var_is_set(var):
        raise ValueError(f"Missing required environment variable {var}.")

    path = os.environ[var]
    ensure_dir_exists(path)
    return path


def get_source_path():
    return get_environment_variable("YUCCA_SOURCE")


def get_raw_data_path():
    return get_environment_variable("YUCCA_RAW_DATA")


def get_preprocessed_data_path():
    return get_environment_variable("YUCCA_PREPROCESSED_DATA")


def get_models_path():
    return get_environment_variable("YUCCA_MODELS")


def get_results_path():
    return get_environment_variable("YUCCA_RESULTS")


def get_yucca_wandb_entity():
    return os.getenv("YUCCA_WANDB_ENTITY")



from yucca.modules.networks.networks import TinyUNet
from yucca.modules.optimization.loss_functions.nnUNet_losses import DiceCE

model = TinyUNet

config = {
    "batch_size": 2,
    "classes": [0, 1],
    "config_name": "demo",
    "crop_to_nonzero": True,
    "continue_from_most_recent": True,
    "deep_supervision": False,
    "experiment": "default",
    "extension": ".nii.gz",
    "learning_rate": 1e-3,
    "loss_fn": DiceCE,
    "max_epochs": 2,
    "modalities": ("MRI",),
    "model_dimensions": "2D",
    "model": TinyUNet,
    "momentum": 0.99,
    "norm_op": "volume_wise_znorm",
    "patch_size": (32, 32),
    "plans": None,
    "split_idx": 0,
    "split_method": "kfold",
    "split_param": 5,
    "target_size": None,
    "target_spacing": [1.0, 1.0, 1.0],
    "target_coordinate_system": "RAS",
    "task": "Task000_TEST_SEGMENTATION",
    "task_type": "segmentation",
}

config["model_name"] = config["model"].__name__
config["num_classes"] = len(config["classes"])
config["num_modalities"] = len(config["modalities"])



import logging

logging.getLogger().setLevel(logging.INFO)


# Tell PyInstaller where to find hook-yucca.py
def _pyinstaller_hooks_dir():
    from pathlib import Path

    return [str(Path(__file__).with_name("_pyinstaller").resolve())]



import torch


def torch_gibbs_ringing(
    image: torch.Tensor, num_sample: int, mode: str = "rect", axes: list[int] = None, clip_to_input_range: bool = False
) -> torch.Tensor:
    assert image.ndim in [2, 3], "Only 2D or 3D images supported"
    if mode == "rect":
        assert axes is not None and all(0 <= ax < image.ndim for ax in axes), f"Invalid axes for mode 'rect'"

    img_min = image.min()
    img_max = image.max()
    offset = -img_min if img_min < 0 else 0
    image = image + offset

    kspace = torch.fft.fftshift(torch.fft.fftn(image, dim=list(range(image.ndim))), dim=list(range(image.ndim)))

    shape = image.shape
    center = [s // 2 for s in shape]

    if mode == "rect":
        mask = torch.ones_like(kspace, dtype=torch.bool)
        for axis in axes:
            c = center[axis]
            half = num_sample // 2
            slc = [slice(None)] * image.ndim
            for i in range(shape[axis]):
                if not (c - half <= i < c + half):
                    slc[axis] = slice(i, i + 1)
                    mask[tuple(slc)] = False
        kspace[~mask] = 0
    elif mode == "radial":
        coords = torch.meshgrid([torch.arange(s, device=image.device) - c for s, c in zip(shape, center)], indexing="ij")
        dist = torch.sqrt(sum((g.float() ** 2 for g in coords)))
        mask = dist <= num_sample
        kspace[~mask] = 0
    else:
        raise ValueError(f"Unsupported mode: {mode}")

    result = torch.fft.ifftn(torch.fft.ifftshift(kspace, dim=list(range(image.ndim))), dim=list(range(image.ndim)))
    result = result.abs()

    result = result - offset
    if clip_to_input_range:
        result = torch.clamp(result, min=img_min, max=img_max)

    return result



from yucca.pipeline.planning.YuccaPlanner import YuccaPlanner
import numpy as np


class YuccaPlannerV2(YuccaPlanner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = str(self.__class__.__name__)
        self.get_foreground_locations_per_label = True

    def determine_norm_op_per_modality(self):
        self.norm_op_per_modality = [
            "ct" if "ct" in mod.lower() else "volume_wise_znorm" for mod in self.dataset_properties["modalities"].values()
        ]

    def determine_transpose(self):
        # If no specific view is determined in run training, we select the optimal.
        # This will be the optimal solution in most cases that are not 2.5D training.
        median_spacing = self.dataset_properties["original_median_spacing"]
        sorting_key = np.argsort(median_spacing)[::-1]
        self.transpose_fw = sorting_key.tolist()
        self.transpose_bw = sorting_key.argsort().tolist()

    def determine_target_size_from_fixed_size_or_spacing(self):
        # If the dataset contains data with very different spacings this planner will
        # not use the median spacing, but rather the first quartile spacing.
        # By doing so it intentionally avoids excessive downsampling and instead favors upsampling.
        self.fixed_target_size = None
        self.fixed_target_spacing = self.dataset_properties["original_median_spacing"]
        spacing_anisotropy_threshold = 2.5
        spacings = np.array(self.dataset_properties["original_spacings"])
        for dim in range(spacings.shape[1]):
            low = np.percentile(spacings[:, dim], 5)
            high = np.percentile(spacings[:, dim], 95)
            if low * spacing_anisotropy_threshold < high:
                self.fixed_target_spacing[dim] = np.percentile(spacings[:, dim], 10)



if __name__ == "__main__":
    import lightning as L
    import os
    import torch
    from batchgenerators.utilities.file_and_folder_operations import maybe_mkdir_p as ensure_dir_exists
    from yucca.paths import (
        get_models_path,
        get_results_path,
        get_preprocessed_data_path,
        get_raw_data_path,
    )
    from yucca.modules.callbacks.prediction_writer import WritePredictionFromLogits
    from yucca.modules.lightning_modules.BaseLightningModule import BaseLightningModule
    from yucca.modules.data.data_modules.YuccaDataModule import YuccaDataModule
    from yucca.modules.data.datasets.YuccaDataset import YuccaTestPreprocessedDataset
    from yucca.pipeline.evaluation.YuccaEvaluator import YuccaEvaluator
    from yucca.documentation.templates.template_config import config

    ckpt_path = os.path.join(
        get_models_path(),
        config["task"],
        config["model_name"] + "__" + config["model_dimensions"],
        "__" + config["config_name"],
        "default",
        "kfold_5_fold_0",
        "version_0",
        "checkpoints",
        "last.ckpt",
    )

    gt_path = os.path.join(get_raw_data_path(), config["task"], "labelsTs")
    target_data_path = os.path.join(get_preprocessed_data_path(), config["task"] + "_test", "demo")

    save_path = os.path.join(
        get_results_path(),
        config["task"],
        config["task"],
        config["model_name"] + "__" + config["model_dimensions"],
        "__" + config["config_name"],
        "kfold_5_fold_0",
        "version_0",
        "best",
    )
    ensure_dir_exists(save_path)

    ckpt = torch.load(ckpt_path, map_location="cpu")
    pred_writer = WritePredictionFromLogits(output_dir=save_path, save_softmax=False, write_interval="batch")

    model_module = BaseLightningModule(
        model=config["model"],
        model_dimensions=config["model_dimensions"],
        num_classes=config["num_classes"],
        num_modalities=config["num_modalities"],
        patch_size=config["patch_size"],
        crop_to_nonzero=config["crop_to_nonzero"],
        disable_inference_preprocessing=True,
    )

    data_module = YuccaDataModule(
        batch_size=config["batch_size"],
        patch_size=config["patch_size"],
        pred_save_dir=save_path,
        pred_data_dir=target_data_path,
        overwrite_predictions=True,
        image_extension=".nii.gz",
        test_dataset_class=YuccaTestPreprocessedDataset,
    )

    trainer = L.Trainer(
        callbacks=pred_writer,
        precision="32",
        enable_progress_bar=True,
    )

    trainer.predict(
        model=model_module,
        dataloaders=data_module,
        ckpt_path=ckpt_path,
        return_predictions=False,
    )

    evaluator = YuccaEvaluator(
        labels=config["classes"],
        folder_with_ground_truth=gt_path,
        folder_with_predictions=save_path,
        use_wandb=False,
    )
    evaluator.run()



from .croppad import torch_croppad
from .bias_field import torch_bias_field
from .blur import torch_blur
from .gamma import torch_gamma
from .motion_ghosting import torch_motion_ghosting
from .masking import torch_mask
from .noise import torch_additive_noise, torch_multiplicative_noise
from .ringing import torch_gibbs_ringing
from .sampling import torch_simulate_lowres
from .spatial import torch_spatial



import numpy as np
from skimage.morphology import skeletonize, dilation


def skeleton(label, do_tube: bool = True):
    # Add tubed skeleton GT
    label_copy = label
    bin_seg = label_copy > 0
    seg_all_skel = np.zeros_like(bin_seg, dtype=np.int16)

    # Skeletonize
    if not np.sum(bin_seg) == 0:
        skel = skeletonize(bin_seg)
        skel = (skel > 0).astype(np.int16)
        if do_tube:
            skel = dilation(dilation(skel))
        skel *= label_copy.astype(np.int16)
    return seg_all_skel



import torch
import torch.nn.functional as F


def torch_simulate_lowres(
    image: torch.Tensor,
    target_shape: tuple[int],
    clip_to_input_range: bool,
) -> torch.Tensor:
    original_shape = image.shape
    img_min = image.min()
    img_max = image.max()

    image = image.float()
    if image.ndim == 2:
        image = image.unsqueeze(0).unsqueeze(0)
        mode_down = "nearest-exact"
        mode_up = "bicubic"
    elif image.ndim == 3:
        image = image.unsqueeze(0).unsqueeze(0)
        mode_down = "nearest-exact"
        mode_up = "trilinear"
    else:
        raise ValueError("Image must be 3D or 4D.")

    downsampled = F.interpolate(image, size=tuple(target_shape), mode=mode_down)
    upsampled = F.interpolate(downsampled, size=tuple(original_shape), mode=mode_up, align_corners=False)
    result = upsampled.squeeze(0).squeeze(0)

    if clip_to_input_range:
        result = result.clamp(min=img_min.item(), max=img_max.item())

    return result



from yucca.modules.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.transforms import skeleton


class Skeleton(YuccaTransform):
    def __init__(self, skeleton=False, label_key="label", do_tube=False):
        self.label_key = label_key
        self.skeleton = skeleton
        self.do_tube = do_tube

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __skeletonize__(self, label):
        label_skeleton = label.copy()
        for b in range(label.shape[0]):
            for c in range(label.shape[1]):
                label_skeleton[b, c] = skeleton(label[b, c])
        return label_skeleton[0]

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        if self.skeleton:
            data_dict["skel"] = self.__skeletonize__(data_dict[self.label_key])
        return data_dict



if __name__ == "__main__":
    import re
    import os
    import numpy as np
    import torch
    from batchgenerators.utilities.file_and_folder_operations import (
        subfiles,
        join,
        save_pickle,
        maybe_mkdir_p as ensure_dir_exists,
    )
    from yucca.paths import get_raw_data_path, get_preprocessed_data_path
    from yucca.documentation.templates.template_config import config
    from yucca.functional.preprocessing import preprocess_case_for_training_with_label, preprocess_case_for_inference
    from yucca.functional.utils.loading import read_file_to_nifti_or_np

    raw_images_dir = join(get_raw_data_path(), config["task"], "imagesTr")
    raw_labels_dir = join(get_raw_data_path(), config["task"], "labelsTr")
    test_raw_images_dir = join(get_raw_data_path(), config["task"], "imagesTs")

    target_dir = join(get_preprocessed_data_path(), config["task"], config["config_name"])
    test_target_dir = join(get_preprocessed_data_path(), config["task"] + "_test", config["config_name"])

    ensure_dir_exists(target_dir)
    ensure_dir_exists(test_target_dir)

    # Preprocess the training data
    subjects = [file[: -len(config["extension"])] for file in subfiles(raw_labels_dir, join=False) if not file.startswith(".")]

    for sub in subjects:
        images = [
            image_path
            for image_path in subfiles(raw_images_dir)
            if re.search(re.escape(sub) + "_" + r"\d{3}" + ".", os.path.split(image_path)[-1])
        ]
        images = [read_file_to_nifti_or_np(image) for image in images]
        label = read_file_to_nifti_or_np(join(raw_labels_dir, sub + config["extension"]))
        images, label, image_props = preprocess_case_for_training_with_label(
            images=images,
            label=label,
            normalization_operation=[config["norm_op"] for _ in config["modalities"]],
            allow_missing_modalities=False,
            enable_cc_analysis=False,
            crop_to_nonzero=config["crop_to_nonzero"],
        )
        images = np.vstack((np.array(images), np.array(label)[np.newaxis]), dtype=np.float32)

        save_path = join(target_dir, sub)
        np.save(save_path + ".npy", images)
        save_pickle(image_props, save_path + ".pkl")

    # Preprocess the test data
    subjects = [
        file[: -len("_000" + config["extension"])]
        for file in subfiles(test_raw_images_dir, join=False)
        if not file.startswith(".")
    ]

    for sub in subjects:
        images = [
            image_path
            for image_path in subfiles(test_raw_images_dir)
            if re.search(re.escape(sub) + "_" + r"\d{3}" + ".", os.path.split(image_path)[-1])
        ]
        images = [read_file_to_nifti_or_np(image) for image in images]
        images, image_props = preprocess_case_for_inference(
            crop_to_nonzero=config["crop_to_nonzero"],
            images=images,
            intensities=None,
            normalization_scheme=[config["norm_op"] for _ in config["modalities"]],
            patch_size=config["patch_size"],
            target_size=config["target_size"],
            target_spacing=config["target_spacing"],
            target_orientation=config["target_coordinate_system"],
        )
        save_path = join(test_target_dir, sub)
        torch.save(images, save_path + ".pt")
        save_pickle(image_props, save_path + ".pkl")



import torch


def torch_gamma(
    image: torch.Tensor,
    gamma_range=(0.5, 2),
    invert_image=False,
    epsilon=1e-7,
    per_channel=False,
    clip_to_input_range=False,
) -> torch.Tensor:
    if invert_image:
        image = -image

    if not per_channel:
        if torch.rand(1).item() < 0.5 and gamma_range[0] < 1:
            gamma = torch.rand(1).item() * (1 - gamma_range[0]) + gamma_range[0]
        else:
            gamma = torch.rand(1).item() * (gamma_range[1] - max(gamma_range[0], 1)) + max(gamma_range[0], 1)

        img_min = image.min()
        img_max = image.max()
        img_range = img_max - img_min

        image = torch.pow(((image - img_min) / (img_range + epsilon)), gamma) * img_range + img_min

        if clip_to_input_range:
            image = torch.clamp(image, min=img_min, max=img_max)
    else:
        for c in range(image.shape[0]):
            if torch.rand(1).item() < 0.5 and gamma_range[0] < 1:
                gamma = torch.rand(1).item() * (1 - gamma_range[0]) + gamma_range[0]
            else:
                gamma = torch.rand(1).item() * (gamma_range[1] - max(gamma_range[0], 1)) + max(gamma_range[0], 1)

            img_min = image[c].min()
            img_max = image[c].max()
            img_range = img_max - img_min

            image[c] = torch.pow(((image[c] - img_min) / (img_range + epsilon)), gamma) * (img_range + epsilon) + img_min

            if clip_to_input_range:
                image[c] = torch.clamp(image[c], min=img_min, max=img_max)

    return image



import torch


def torch_motion_ghosting(
    image: torch.Tensor,
    alpha: float,
    num_reps: int,
    axis: int,
    clip_to_input_range: bool = False,
) -> torch.Tensor:
    img_min = image.min()
    img_max = image.max()
    m = min(0, img_min.item())
    image = image + abs(m)

    if image.ndim == 3:
        image = torch.fft.fftn(image, dim=[0, 1, 2])

        if axis == 0:
            image[0:-1:num_reps, :, :] = alpha * image[0:-1:num_reps, :, :]
        elif axis == 1:
            image[:, 0:-1:num_reps, :] = alpha * image[:, 0:-1:num_reps, :]
        else:
            image[:, :, 0:-1:num_reps] = alpha * image[:, :, 0:-1:num_reps]

        image = torch.abs(torch.fft.ifftn(image, dim=[0, 1, 2]))

    elif image.ndim == 2:
        image = torch.fft.fftn(image, dim=[0, 1])

        if axis == 0:
            image[0:-1:num_reps, :] = alpha * image[0:-1:num_reps, :]
        else:
            image[:, 0:-1:num_reps] = alpha * image[:, 0:-1:num_reps]

        image = torch.abs(torch.fft.ifftn(image, dim=[0, 1]))

    image = image - abs(m)

    if clip_to_input_range:
        image = torch.clamp(image, min=img_min, max=img_max)

    return image



import nibabel as nib
import numpy as np
from yucca.functional.utils.softmax import softmax
from yucca.functional.utils.nib_utils import reorient_nib_image
from PIL import Image
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    subfiles,
    maybe_mkdir_p as ensure_dir_exists,
)


def save_nifti_from_numpy(pred, outpath, properties, compression=9):
    # slight hacky, but it is what it is
    nib.openers.Opener.default_compresslevel = compression
    pred = nib.Nifti1Image(
        pred,
        properties["nifti_metadata"]["affine"],
        header=nib.Nifti1Header(properties["nifti_metadata"]["header"]),
        dtype=np.uint8,
    )
    if properties["nifti_metadata"]["reoriented"]:
        pred = reorient_nib_image(
            pred, properties["nifti_metadata"]["final_direction"], properties["nifti_metadata"]["original_orientation"]
        )
    nib.save(
        pred,
        outpath + ".nii.gz",
    )
    del pred


def save_png_from_numpy(pred, outpath):
    pred = Image.fromarray(pred.astype(np.uint8))
    pred.save(outpath + ".png")
    del pred


def save_txt_from_numpy(pred, outpath):
    np.savetxt(outpath + ".txt", np.atleast_1d(pred), fmt="%i", delimiter=",")
    del pred


def save_prediction_from_logits(logits, outpath, properties, save_softmax=False, compression=9):
    if save_softmax:
        softmax_result = softmax(logits)[0].astype(np.float32)
        np.savez_compressed(outpath + ".npz", data=softmax_result, properties=properties)
    if logits.shape[1] > 1:
        logits = np.argmax(logits, 1)
    pred = np.squeeze(logits)
    if properties.get("save_format") == "png":
        save_png_from_numpy(pred, outpath)
    elif properties.get("save_format") == "txt":
        save_txt_from_numpy(pred, outpath)
    else:
        save_nifti_from_numpy(pred, outpath, properties, compression=compression)


def save_multilabel_prediction_from_logits(logits, outpath, properties, compression=9):
    nib.openers.Opener.default_compresslevel = compression
    # here we use a sigmoid instead of the softmax/argmax functions to keep the multiple channels
    # of predicted classes. This means predictions from this are stored as (c, h, w, d) rather than (h, w, d)
    pred = 1.0 / (1.0 + np.exp(-logits))
    pred = (pred > 0.5).astype(np.uint8)
    pred = np.squeeze(pred)
    preds = []
    for i in range(pred.shape[0]):
        pred_for_label = nib.Nifti1Image(
            pred[i],
            properties["nifti_metadata"]["affine"],
            header=nib.Nifti1Header(properties["nifti_metadata"]["header"]),
            dtype=np.uint8,
        )
        if properties["nifti_metadata"]["reoriented"]:
            pred_for_label = reorient_nib_image(
                pred_for_label,
                properties["nifti_metadata"]["final_direction"],
                properties["nifti_metadata"]["original_orientation"],
            )
        preds.append(pred_for_label)
    pred = nib.concat_images(preds)
    nib.save(
        pred,
        outpath + ".nii.gz",
    )
    del pred


def merge_softmax_from_folders(folders: list, outpath, method="sum"):
    ensure_dir_exists(outpath)
    cases = subfiles(folders[0], suffix=".npz", join=False)
    for folder in folders:
        assert cases == subfiles(folder, suffix=".npz", join=False), (
            f"Found unexpected cases. "
            f"The following two folders do not contain the same cases: \n"
            f"{folders[0]} \n"
            f"{folder}"
        )

    for case in cases:
        files_for_case = [np.load(join(folder, case), allow_pickle=True) for folder in folders]
        properties_for_case = files_for_case[0]["properties"]
        files_for_case = [file["data"].astype(np.float32) for file in files_for_case]

        if method == "sum":
            files_for_case = np.sum(files_for_case, axis=0)

        files_for_case = np.argmax(files_for_case, 0)
        save_nifti_from_numpy(
            files_for_case,
            join(outpath, case[:-4]),
            properties=properties_for_case.item(),
        )

    del files_for_case, properties_for_case



#! /usr/bin/env python3

# How to run:
# python viz_torch_transforms --datasets-dir /path/to/datasets --output /path/to/output.png
# python viz_torch_transforms --datasets-dir /path/to/datasets --output /path/to/output.png --headless
# Defaults: datasets dir is ~/datasets and output is ./viz_torch_transforms.png

import os
import glob
import nibabel as nib
import matplotlib.pyplot as plt
import argparse
import torch
import numpy as np

from yucca.functional.transforms.torch import (
    torch_bias_field,
    torch_blur,
    torch_gamma,
    torch_motion_ghosting,
    torch_mask,
    torch_additive_noise,
    torch_multiplicative_noise,
    torch_gibbs_ringing,
    torch_simulate_lowres,
    torch_spatial,
)

torch.manual_seed(1234)


def show_high_res(data, title):
    """Show a high-resolution view of the data with slice navigation."""
    print(f"\nOpening high-resolution view for: {title}")
    print(f"Data shape: {data.shape}")

    fig, ax = plt.subplots(figsize=(10, 10))
    fig.canvas.manager.set_window_title(f"High Resolution: {title}")
    current_slice = data.shape[2] // 2

    # Show initial image and colorbar
    im = ax.imshow(data[:, :, current_slice], cmap="gray")
    ax.set_title(f"{title}\nSlice {current_slice} of {data.shape[2]}", pad=20)
    # cbar = plt.colorbar(im, ax=ax, label='Intensity')

    def update_display():
        im.set_data(data[:, :, current_slice])
        ax.set_title(f"{title}\nSlice {current_slice} of {data.shape[2]}", pad=20)
        # Optionally, update colorbar limits if data range changes
        # im.set_clim(vmin=data.min(), vmax=data.max())
        fig.canvas.draw_idle()

    def on_key(event):
        nonlocal current_slice
        print(f"Key pressed: {event.key}")
        if event.key == "up" and current_slice < data.shape[2] - 1:
            current_slice += 1
            print(f"Moving to slice {current_slice}")
            update_display()
        elif event.key == "down" and current_slice > 0:
            current_slice -= 1
            print(f"Moving to slice {current_slice}")
            update_display()

    fig.canvas.mpl_connect("key_press_event", on_key)
    plt.show()


def on_click(event):
    """Handle click events on the main figure."""
    print(f"\nClick event detected at: {event.xdata}, {event.ydata}")

    if event.inaxes is None:
        print("Click was outside axes")
        return

    ax = event.inaxes
    print(f"Clicked on axes: {ax}")

    for idx, (title, data) in enumerate(transforms.items()):
        if ax == axes[idx // n_cols, idx % n_cols]:
            print(f"Found matching subplot: {title}")
            show_high_res(data.numpy(), title)
            return

    print("No matching subplot found")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Apply and visualize image transformations")
    parser.add_argument("--headless", action="store_true", help="Run without displaying plots")
    parser.add_argument(
        "--datasets-dir",
        type=str,
        default=os.path.expanduser("~/datasets"),
        help="Path to the datasets directory (default: ~/datasets)",
    )
    parser.add_argument(
        "--output",
        type=str,
        default="viz_torch_transforms.png",
        help="Output image file path (default: viz_torch_transforms.png)",
    )
    args = parser.parse_args()

    nii_files = glob.glob(os.path.join(args.datasets_dir, "**", "*.nii.gz"), recursive=True)
    nii_files.sort()

    single_sample = nii_files[1]
    im = nib.load(single_sample)
    data = im.get_fdata()

    print("\nImage Analysis:")
    print("Data type:", data.dtype)
    print("Shape:", data.shape)
    print("Value range:", data.min(), "to", data.max())
    print("Mean value:", data.mean())
    print("Std value:", data.std())
    print("Unique values count:", len(np.unique(data)))
    print("First few unique values:", np.unique(data)[:10])

    # Convert to float32 and scale to [0,1] while preserving relative intensities
    imarr = data.astype(np.float32)
    data_min = imarr.min()
    data_max = imarr.max()
    data_range = data_max - data_min

    # Scale to [0,1] while preserving relative intensities
    imarr = (imarr - data_min) / data_range
    imarr = torch.from_numpy(imarr)
    # Assume batch and channel dimensions are present
    # imarr = imarr.unsqueeze(0).unsqueeze(0)

    print("\nNormalized Image Analysis:")
    print("Value range:", imarr.min().item(), "to", imarr.max().item())
    print("Mean value:", imarr.mean().item())
    print("Std value:", imarr.std().item())

    transforms = {
        "Original": imarr,
        "Bias Field": torch_bias_field(imarr.clone(), clip_to_input_range=True),
        "Blurred (σ=2.0)": torch_blur(imarr.clone(), sigma=2.0),
        "Gamma (0.5–2.0)": torch_gamma(imarr.clone(), gamma_range=(0.5, 2.0), clip_to_input_range=True),
        "Gamma (2.0)": torch_gamma(imarr, gamma_range=(2.0, 2.0), clip_to_input_range=True),
        "Ghost (α=2, ax=0)": torch_motion_ghosting(imarr.clone(), alpha=2.0, num_reps=4, axis=0, clip_to_input_range=True),
        "Ghost (α=2, ax=1)": torch_motion_ghosting(imarr.clone(), alpha=2.0, num_reps=4, axis=1, clip_to_input_range=True),
        "Masked (r=0.3)": torch_mask(imarr.clone(), pixel_value=0, ratio=0.3, token_size=[16, 16, 16]),
        "Add Noise (σ=0.05)": torch_additive_noise(imarr.clone(), mean=0, sigma=0.05, clip_to_input_range=True),
        "Add Noise (σ=0.1)": torch_additive_noise(imarr.clone(), mean=0, sigma=0.1, clip_to_input_range=True),
        "Add Noise (σ=0.2)": torch_additive_noise(imarr.clone(), mean=0, sigma=0.2, clip_to_input_range=True),
        "Mult Noise (σ=0.05)": torch_multiplicative_noise(imarr.clone(), mean=0, sigma=0.05, clip_to_input_range=True),
        "Mult Noise (σ=0.1)": torch_multiplicative_noise(imarr.clone(), mean=0, sigma=0.1, clip_to_input_range=True),
        "Mult Noise (σ=0.2)": torch_multiplicative_noise(imarr.clone(), mean=0, sigma=0.2, clip_to_input_range=True),
        "Gibbs (axes=0,1,2)": torch_gibbs_ringing(
            imarr.clone(), num_sample=64, axes=[0, 1, 2], mode="rect", clip_to_input_range=True
        ),
        "Low Res": torch_simulate_lowres(imarr.clone(), target_shape=(32, 32, 32), clip_to_input_range=True),
        "Spatial": torch_spatial(
            imarr.clone(),
            patch_size=imarr.shape,
            p_deform=1.0,
            p_rot=1.0,
            p_rot_per_axis=1.0,
            p_scale=1.0,
            alpha=10.0,
            sigma=3.0,
            x_rot=0.5,
            y_rot=0,
            z_rot=0,
            scale_factor=1.0,
            clip_to_input_range=True,
        )[
            0
        ],  # Get only the image, not the label
    }

    # Get middle slice for visualization
    slice_idx = imarr.shape[-1] // 2  # Changed to use last dimension

    # Calculate grid dimensions
    n_transforms = len(transforms)
    n_cols = 6  # Show 6 images per row
    n_rows = (n_transforms + n_cols - 1) // n_cols  # Ceiling division

    # Create a figure with subplots
    fig_width = 2.2 * n_cols
    fig_height = 2.2 * n_rows
    plt.rcParams["figure.figsize"] = [fig_width, fig_height]
    plt.rcParams["figure.dpi"] = 200  # Keep high DPI for quality
    fig, axes = plt.subplots(n_rows, n_cols, squeeze=False)

    # Plot each transformation
    for idx, (title, transformed_data) in enumerate(transforms.items()):
        row = idx // n_cols
        col = idx % n_cols

        # Plot transformed image
        if isinstance(transformed_data, tuple):
            transformed_data = transformed_data[0]
        print(title, transformed_data.shape)
        transformed_plot = transformed_data.numpy()
        axes[row, col].imshow(transformed_plot[:, :, slice_idx], cmap="gray")
        axes[row, col].set_title(title, pad=1, fontsize=7)
        axes[row, col].axis("off")

    # Hide any unused subplots
    for idx in range(len(transforms), n_rows * n_cols):
        row = idx // n_cols
        col = idx % n_cols
        axes[row, col].axis("off")
        axes[row, col].set_visible(False)

    # Use subplots_adjust for tight packing
    plt.subplots_adjust(top=0.92, bottom=0.08, left=0.03, right=0.97, wspace=0.02, hspace=0.15)

    # Set window title
    fig.canvas.manager.set_window_title("Image Transformations")

    # Save the figure
    plt.savefig(args.output, dpi=200, bbox_inches="tight")
    print(f"\nSaved visualization to: {args.output}")

    # Connect the click event
    print("\nConnecting click event handler...")
    fig.canvas.mpl_connect("button_press_event", on_click)
    print("Click event handler connected. Click any image to view in high resolution.")

    # Only show the plot if not in headless mode
    if not args.headless:
        plt.show()



import torch


class SingleFileTestDataset(torch.utils.data.Dataset):
    def __init__(self, raw_data_dir: str, pred_save_dir: str, suffix=".nii.gz", **kwargs):  # noqa: U100
        self.data_path = raw_data_dir
        self.pred_save_dir = pred_save_dir
        self.suffix = suffix

    def __len__(self):
        return 1

    def __getitem__(self, idx):  # noqa: U100
        # Here we generate the paths to the cases along with their ID which they will be saved as.
        # we pass "case" as a list of strings and case_id as a string to the dataloader which
        # will convert them to a list of tuples of strings and a tuple of a string.
        # i.e. ['path1', 'path2'] -> [('path1',), ('path2',)]

        return {"data_paths": [self.data_path], "extension": self.suffix, "case_id": ""}



from yucca.modules.data.augmentation.transforms.BiasField import BiasField, Torch_BiasField
from yucca.modules.data.augmentation.transforms.Blur import Blur, Torch_Blur
from yucca.modules.data.augmentation.transforms.convert_labels_to_regions import ConvertLabelsToRegions
from yucca.modules.data.augmentation.transforms.copy_image_to_label import CopyImageToLabel, Torch_CopyImageToLabel
from yucca.modules.data.augmentation.transforms.cropping_and_padding import CropPad, Torch_CropPad
from yucca.modules.data.augmentation.transforms.formatting import (
    RemoveBatchDimension,
    RemoveSegChannelAxis,
    NumpyToTorch,
    AddBatchDimension,
    CollectMetadata,
)
from yucca.modules.data.augmentation.transforms.Gamma import Gamma, Torch_Gamma
from yucca.modules.data.augmentation.transforms.Ghosting import MotionGhosting, Torch_MotionGhosting
from yucca.modules.data.augmentation.transforms.Masking import Masking, Torch_Mask
from yucca.modules.data.augmentation.transforms.Mirror import Mirror
from yucca.modules.data.augmentation.transforms.Noise import (
    AdditiveNoise,
    MultiplicativeNoise,
    Torch_AdditiveNoise,
    Torch_MultiplicativeNoise,
)
from yucca.modules.data.augmentation.transforms.normalize import Normalize
from yucca.modules.data.augmentation.transforms.Ringing import GibbsRinging, Torch_GibbsRinging
from yucca.modules.data.augmentation.transforms.sampling import DownsampleSegForDS
from yucca.modules.data.augmentation.transforms.SimulateLowres import SimulateLowres, Torch_SimulateLowres
from yucca.modules.data.augmentation.transforms.Spatial import Spatial, Torch_Spatial
from yucca.modules.data.augmentation.transforms.Skeleton import Skeleton



import os
from typing import Literal, Optional, Union
import numpy as np
import json
import wandb
from batchgenerators.utilities.file_and_folder_operations import subfiles, join, load_json, isfile
from yucca.functional.evaluation.metrics import (
    dice,
    jaccard,
    sensitivity,
    precision,
    TP,
    FP,
    FN,
    total_pos_gt,
    total_pos_pred,
    volume_similarity,
    accuracy,
)
from yucca.functional.evaluation.evaluate_folder import (
    evaluate_folder_segm,
    evaluate_folder_cls,
)
from yucca.paths import get_raw_data_path


class YuccaEvaluator(object):
    def __init__(
        self,
        labels: Union[dict, int],
        folder_with_predictions,
        folder_with_ground_truth,
        use_wandb: bool,
        as_binary=False,
        do_object_eval=False,
        do_surface_eval=False,
        regions: Optional[dict] = None,
        overwrite: bool = False,
        surface_tol: int = 1,
        task_type: Literal["segmentation", "classification", "regression"] = "segmentation",
        extension: str = None,
        strict: bool = True,
    ):
        self.name = "results"

        self.labels = labels
        self.regions = regions
        self.overwrite = overwrite
        self.use_wandb = use_wandb
        self.task_type = task_type
        self.strict = strict
        self.as_binary = as_binary
        self.surface_tol = surface_tol

        self.metrics = {
            "Dice": dice,
            "Jaccard": jaccard,
            "Sensitivity": sensitivity,
            "Precision": precision,
            "Volume Similarity": volume_similarity,
            "True Positives": TP,
            "False Positives": FP,
            "False Negatives": FN,
            "Total Positives Ground Truth": total_pos_gt,
            "Total Positives Prediction": total_pos_pred,
        }
        self.obj_metrics = []
        self.surface_metrics = []

        if self.task_type == "segmentation":
            self.metrics = {
                "Dice": dice,
                "Jaccard": jaccard,
                "Sensitivity": sensitivity,
                "Precision": precision,
                "Volume Similarity": volume_similarity,
                "True Positives": TP,
                "False Positives": FP,
                "False Negatives": FN,
                "Total Positives Ground Truth": total_pos_gt,
                "Total Positives Prediction": total_pos_pred,
            }

            if do_object_eval:
                self.name += "_OBJ"
                self.obj_metrics = [
                    "_OBJ Total Objects Prediction",
                    "_OBJ Total Objects Ground Truth",
                    "_OBJ True Positives",
                    "_OBJ False Positives",
                    "_OBJ False Negatives",
                    "_OBJ Mean Volume Prediction",
                    "_OBJ Mean Volume Ground Truth",
                    "_OBJ sensitivity",
                    "_OBJ precision",
                    "_OBJ F1",
                ]

            if do_surface_eval:
                self.name += f"_SURFACE{self.surface_tol}"
                self.surface_metrics = [
                    "Average Surface Distance",
                ]

            self.metrics_included_in_streamtable = [
                "Dice",
                "Sensitivity",
                "Precision",
                "Jaccard",
                "Volume Similarity",
                "_OBJ sensitivity",
                "_OBJ precision",
                "_OBJ F1",
            ]
        elif self.task_type == "classification":
            self.metrics = {
                "Accuracy": accuracy,
                "F1": dice,
                "Sensitivity": sensitivity,
                "Precision": precision,
                # "AUROC": auroc, # only when probabilities are available
            }

            self.metrics_included_in_streamtable = [
                "Accuracy",
                "F1",
                "Sensitivity",
                "Precision",
                # "AUROC", # only when probabilities are available
            ]
        elif self.task_type == "regression":
            raise NotImplementedError
            # self.metrics = {
            #     "MAE": mae,
            #     "RMSE": rmse,
            #     "R2": r2,
            # }
        else:
            raise ValueError(f"Unknown task type {self.task_type}")

        if self.as_binary:
            self.name += "_BINARY"
            self.labels = [0, 1]
            self.labelarr = np.sort(np.array(self.labels, dtype=np.uint8))
        elif isinstance(self.labels, int):
            self.labelarr = np.sort(np.arange(self.labels, dtype=np.uint8))
        elif isinstance(self.labels, list):
            self.labelarr = np.sort(np.array(self.labels, dtype=np.uint8))
        elif isinstance(self.labels, dict):
            self.labelarr = np.sort(np.array(list(self.labels.keys()), dtype=np.uint8))
        else:
            raise ValueError(f"Incorrect label format. Got {type(self.labels)}")

        self.folder_with_predictions = folder_with_predictions
        self.folder_with_ground_truth = folder_with_ground_truth

        self.outpath = join(self.folder_with_predictions, f"{self.name}.json")

        if extension is None:
            if self.task_type == "classification":
                extension = ".txt"
            else:
                extension = "nii.gz"

        self.pred_subjects = subfiles(self.folder_with_predictions, suffix=extension, join=False)
        self.gt_subjects = subfiles(self.folder_with_ground_truth, suffix=extension, join=False)

        print(
            f"\n"
            f"STARTING EVALUATION \n"
            f"Looking for {extension} files in the following directiories\n"
            f"Folder with predictions: {self.folder_with_predictions}\n"
            f"Folder with ground truth: {self.folder_with_ground_truth}\n"
        )

    def sanity_checks(self):
        assert self.pred_subjects <= self.gt_subjects, "Ground Truth is missing for predicted scans"

        assert self.gt_subjects <= self.pred_subjects, "Prediction is missing for Ground Truth of scans"

        # Check if the Ground Truth directory is a subdirectory of a 'TaskXXX_MyTask' folder.
        # If so, there should be a dataset.json where we can double check that the supplied classes
        # match with the expected classes for the dataset.
        gt_is_task = [i for i in self.folder_with_ground_truth.split(os.sep) if "Task" in i]
        if gt_is_task:
            gt_task = gt_is_task[0]
            dataset_json = join(get_raw_data_path(), gt_task, "dataset.json")
            if isfile(dataset_json):
                dataset_json = load_json(dataset_json)
                print(f"Labels found in dataset.json: {list(dataset_json['labels'].keys())}")

    def run(self):
        if isfile(self.outpath) and not self.overwrite:
            print(f"Evaluation file already present in {self.outpath}. Skipping.")
        else:
            if self.strict:
                self.sanity_checks()
            results_dict = self.evaluate_folder()
            self.save_as_json(results_dict)
            if self.use_wandb:
                self.update_streamtable(results_dict)

    def evaluate_folder(self):
        if self.task_type == "classification":
            return evaluate_folder_cls(
                labels=self.labelarr,
                metrics=self.metrics,
                subjects=self.pred_subjects,
                folder_with_predictions=self.folder_with_predictions,
                folder_with_ground_truth=self.folder_with_ground_truth,
            )
        elif self.task_type == "segmentation":
            multilabel = self.regions is not None
            return evaluate_folder_segm(
                labels=self.labels if multilabel else self.labelarr,
                metrics=self.metrics,
                subjects=self.pred_subjects,
                folder_with_predictions=self.folder_with_predictions,
                folder_with_ground_truth=self.folder_with_ground_truth,
                as_binary=self.as_binary,
                obj_metrics=self.obj_metrics,
                surface_metrics=self.surface_metrics,
                surface_tol=self.surface_tol,
                regions=self.regions,
                multilabel=multilabel,
            )
        else:
            raise NotImplementedError("Invalid task type")

    def save_as_json(self, dict):
        print("Saving results.json \n \n ########################################################################")
        with open(self.outpath, "w") as f:
            json.dump(dict, f, default=float, indent=4)

    def update_streamtable(self, results_dict):
        """
        Save evaluation results to a wandb StreamTable

        :param results_dict: dictionary with evaluation results
        """
        from weave.monitoring import StreamTable

        task = self.outpath.split(os.path.sep)[-5]
        target = self.outpath.split(os.path.sep)[-6]
        model_name = "/".join(self.outpath.split(os.path.sep)[-4:])

        st = StreamTable(table_name=task, entity_name=wandb.api.viewer()["entity"], project_name="Yucca")

        stream_dict = {"0. Experiment": model_name, "0. Target Task": target}

        if self.task_type == "classification":
            stream_dict = {**stream_dict, **results_dict}

        elif self.task_type == "segmentation":
            for key, _ in results_dict.items():
                if key == "0":
                    continue
                else:
                    stream_dict.update(
                        {f"{key}. " + k: v for k, v in results_dict[key].items() if k in self.metrics_included_in_streamtable}
                    )

        else:
            raise NotImplementedError("Task type not supported")

        st.log(stream_dict)
        st.finish()
